'use strict';

Object.defineProperty(exports, '__esModule', { value: true });

function _interopDefault (ex) { return (ex && (typeof ex === 'object') && 'default' in ex) ? ex['default'] : ex; }

var _defineProperty = _interopDefault(require('@babel/runtime/helpers/defineProperty'));
var _toConsumableArray = _interopDefault(require('@babel/runtime/helpers/toConsumableArray'));
var _slicedToArray = _interopDefault(require('@babel/runtime/helpers/slicedToArray'));
var redux = require('redux');
var _regeneratorRuntime = _interopDefault(require('@babel/runtime/regenerator'));
var _asyncToGenerator = _interopDefault(require('@babel/runtime/helpers/asyncToGenerator'));
var _typeof = _interopDefault(require('@babel/runtime/helpers/typeof'));

function ownKeys(object, enumerableOnly) { var keys = Object.keys(object); if (Object.getOwnPropertySymbols) { var symbols = Object.getOwnPropertySymbols(object); if (enumerableOnly) symbols = symbols.filter(function (sym) { return Object.getOwnPropertyDescriptor(object, sym).enumerable; }); keys.push.apply(keys, symbols); } return keys; }

function _objectSpread(target) { for (var i = 1; i < arguments.length; i++) { var source = arguments[i] != null ? arguments[i] : {}; if (i % 2) { ownKeys(source, true).forEach(function (key) { _defineProperty(target, key, source[key]); }); } else if (Object.getOwnPropertyDescriptors) { Object.defineProperties(target, Object.getOwnPropertyDescriptors(source)); } else { ownKeys(source).forEach(function (key) { Object.defineProperty(target, key, Object.getOwnPropertyDescriptor(source, key)); }); } } return target; }

/**
 * Create the reducer for a model
 *
 * @param {string} name
 * @param {object} state
 * @return {function} Reducer
 */
var createReducer = function createReducer(name, state) {
  state.loading = {};
  return function () {
    var currentState = arguments.length > 0 && arguments[0] !== undefined ? arguments[0] : state;
    var action = arguments.length > 1 ? arguments[1] : undefined;

    var _action$type$split = action.type.split('/'),
        _action$type$split2 = _slicedToArray(_action$type$split, 1),
        modelName = _action$type$split2[0];

    if (modelName !== name) return currentState;
    return _objectSpread({}, currentState, {}, action.payload);
  };
};

/**
 * Get the error types
 */
var ERR = {
  // common
  NOT_STRING: function NOT_STRING(name) {
    return "'".concat(name, "' must be a string");
  },
  NOT_BOOLEAN: function NOT_BOOLEAN(name) {
    return "'".concat(name, "' must be a boolean");
  },
  NOT_ARRAY: function NOT_ARRAY(name) {
    return "'".concat(name, "' must be an array");
  },
  NOT_OBJECT: function NOT_OBJECT(name) {
    return "'".concat(name, "' must be an object");
  },
  // createStore
  DISPATCH: function DISPATCH() {
    return "Please do not use 'dispatch' directly in Retalk";
  },
  MODEL_NAME: function MODEL_NAME(name) {
    return "Illegal model name '".concat(name, "', duplicated with an action name");
  },
  // createReducer
  MODEL: function MODEL(name) {
    return "'".concat(name, "' model must be { state: Object, actions: Object }");
  },
  // createActions
  ACTION_NAME: function ACTION_NAME(name, action) {
    return "Illegal action name '".concat(action, "' in ").concat(name, " model");
  },
  ACTION: function ACTION(name, action) {
    return "Illegal action '".concat(action, "' in ").concat(name, " model");
  },
  // checkDuplicate
  DUPLICATE: function DUPLICATE(name, type, key) {
    return "withStore: Duplicate '".concat(key, "' ").concat(type, " in ").concat(name, " and another model");
  },
  // withStore
  EMPTY_PARAM: function EMPTY_PARAM() {
    return 'Empty model name passed to withStore';
  },
  PARAM: function PARAM() {
    return "Illegal model name passed to withStore";
  }
};
/**
 * Detect an object
 *
 * @param {any} obj
 * @return {boolean}
 */

var isObject = function isObject(obj) {
  return _typeof(obj) === 'object' && obj !== null && !Array.isArray(obj);
};
/**
 * Detect an async function
 *
 * @param {any} fn
 * @return {boolean}
 */

var isAsyncFn = function isAsyncFn(fn) {
  if (typeof fn !== 'function') return false;
  var str = fn.toString();
  return str.includes('regeneratorRuntime.mark(') || str.includes('_regenerator') || str.includes('.apply(');
};
/**
 * Check a model
 *
 * @param {string} name
 * @param {object} model
 */

var checkModel = function checkModel(name, model) {
  if (typeof name !== 'string') throw new Error(ERR.NOT_STRING('name'));
  if (!isObject(model)) throw new Error(ERR.NOT_OBJECT('model'));
  var state = model.state,
      actions = model.actions;

  if (!(Object.keys(model).length === 2 && isObject(state) && isObject(actions))) {
    throw new Error(ERR.MODEL(name));
  }
};
/**
 * Create the handler for a proxy
 *
 * @param {string} name
 * @param {function} getState
 * @param {function} dispatch
 * @param {object} [modelsProxy]
 * @return {object} handler
 */

var createHandler = function createHandler(name, getState, dispatch, modelsProxy) {
  return {
    get: function get(target, prop) {
      if (prop === 'state') return getState()[name];
      if (prop in target) return target[prop];

      if (prop in dispatch[name]) {
        Object.assign(target, dispatch[name]);
        return target[prop];
      }

      if (typeof modelsProxy !== 'undefined' && prop in modelsProxy) {
        Object.assign(target, modelsProxy);
        return target[prop];
      }

      return undefined;
    }
  };
};
/**
 * Check duplicate state or action keys within withStore
 *
 * @param {string} name
 * @param {string} type
 * @param {object} part
 * @param {object} all
 */

var checkDuplicate = function checkDuplicate(name, type, part, all) {
  var keys = Object.keys(part);
  if (type === 'state') keys = keys.filter(function (key) {
    return key !== 'loading';
  });
  keys.forEach(function (key) {
    if (key in all) throw new Error(ERR.DUPLICATE(name, type, key));
  });
};

function ownKeys$1(object, enumerableOnly) { var keys = Object.keys(object); if (Object.getOwnPropertySymbols) { var symbols = Object.getOwnPropertySymbols(object); if (enumerableOnly) symbols = symbols.filter(function (sym) { return Object.getOwnPropertyDescriptor(object, sym).enumerable; }); keys.push.apply(keys, symbols); } return keys; }

function _objectSpread$1(target) { for (var i = 1; i < arguments.length; i++) { var source = arguments[i] != null ? arguments[i] : {}; if (i % 2) { ownKeys$1(source, true).forEach(function (key) { _defineProperty(target, key, source[key]); }); } else if (Object.getOwnPropertyDescriptors) { Object.defineProperties(target, Object.getOwnPropertyDescriptors(source)); } else { ownKeys$1(source).forEach(function (key) { Object.defineProperty(target, key, Object.getOwnPropertyDescriptor(source, key)); }); } } return target; }
/**
 * Format the actions of a model
 *
 * @param {string} name
 * @param {object} model
 * @param {function} getState
 * @param {function} dispatch
 * @param {function} theRealDispatch
 * @param {object} thisProxy
 */

var createActions = function createActions(name, model, getState, dispatch, theRealDispatch, thisProxy) {
  var isEnvDevelopment = process.env.NODE_ENV === 'development';
  var state = model.state,
      actions = model.actions;
  var newActions = {};

  var formatAction = function formatAction(actionName, oldAction) {
    var setState = function reducer(payload) {
      return theRealDispatch({
        type: "".concat(name, "/").concat(actionName),
        payload: payload
      });
    };

    var newAction = function action() {
      thisProxy.setState = setState;
      return oldAction.bind(thisProxy).apply(void 0, arguments);
    };

    if (!isAsyncFn(oldAction)) {
      newActions[actionName] = newAction;
    } else {
      state.loading[actionName] = false;

      var setLoading = function reducer(actionName, actionLoading) {
        return theRealDispatch({
          type: "".concat(name, "/").concat(actionName, "/SET_LOADING"),
          payload: {
            loading: _objectSpread$1({}, getState()[name].loading, _defineProperty({}, actionName, actionLoading))
          }
        });
      };

      newActions[actionName] =
      /*#__PURE__*/
      function () {
        var _asyncAction = _asyncToGenerator(
        /*#__PURE__*/
        _regeneratorRuntime.mark(function _callee() {
          var result,
              _args = arguments;
          return _regeneratorRuntime.wrap(function _callee$(_context) {
            while (1) {
              switch (_context.prev = _context.next) {
                case 0:
                  setLoading(actionName, true);
                  _context.next = 3;
                  return newAction.apply(void 0, _args);

                case 3:
                  result = _context.sent;
                  setLoading(actionName, false);
                  return _context.abrupt("return", result);

                case 6:
                case "end":
                  return _context.stop();
              }
            }
          }, _callee);
        }));

        function asyncAction() {
          return _asyncAction.apply(this, arguments);
        }

        return asyncAction;
      }();
    }
  };

  if (isEnvDevelopment) {
    var forbiddenNames = ['state', 'setState'].concat(_toConsumableArray(Object.keys(state)));
    Object.entries(actions).forEach(function (_ref) {
      var _ref2 = _slicedToArray(_ref, 2),
          actionName = _ref2[0],
          oldAction = _ref2[1];

      if (forbiddenNames.includes(actionName)) throw new Error(ERR.ACTION_NAME(name, actionName));
      if (typeof oldAction !== 'function') throw new Error(ERR.ACTION(name, actionName));
      formatAction(actionName, oldAction);
    });
  } else {
    Object.entries(actions).forEach(function (_ref3) {
      var _ref4 = _slicedToArray(_ref3, 2),
          actionName = _ref4[0],
          oldAction = _ref4[1];

      formatAction(actionName, oldAction);
    });
  }

  dispatch[name] = newActions;
};

function ownKeys$2(object, enumerableOnly) { var keys = Object.keys(object); if (Object.getOwnPropertySymbols) { var symbols = Object.getOwnPropertySymbols(object); if (enumerableOnly) symbols = symbols.filter(function (sym) { return Object.getOwnPropertyDescriptor(object, sym).enumerable; }); keys.push.apply(keys, symbols); } return keys; }

function _objectSpread$2(target) { for (var i = 1; i < arguments.length; i++) { var source = arguments[i] != null ? arguments[i] : {}; if (i % 2) { ownKeys$2(source, true).forEach(function (key) { _defineProperty(target, key, source[key]); }); } else if (Object.getOwnPropertyDescriptors) { Object.defineProperties(target, Object.getOwnPropertyDescriptors(source)); } else { ownKeys$2(source).forEach(function (key) { Object.defineProperty(target, key, Object.getOwnPropertyDescriptor(source, key)); }); } } return target; }
/**
 * Create the store
 *
 * @param {object} models
 * @param {object} options
 * @return {object} Store
 */

var createStore = function createStore(models) {
  var options = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : {};
  var isEnvDevelopment = process.env.NODE_ENV === 'development';
  var _options$useDevTools = options.useDevTools,
      useDevTools = _options$useDevTools === void 0 ? true : _options$useDevTools,
      _options$plugins = options.plugins,
      plugins = _options$plugins === void 0 ? [] : _options$plugins;
  var modelEntries = [];
  var actionNames = [];
  var rootReducers = {};

  if (isEnvDevelopment) {
    // Check the arguments
    if (!isObject(models)) throw new Error(ERR.NOT_OBJECT('models'));
    if (!isObject(options)) throw new Error(ERR.NOT_OBJECT('options'));
    if (typeof useDevTools !== 'undefined' && typeof useDevTools !== 'boolean') throw new Error(ERR.NOT_BOOLEAN('options.useDevTools'));
    if (typeof plugins !== 'undefined' && !Array.isArray(plugins)) throw new Error(ERR.NOT_ARRAY('options.plugins'));
    modelEntries = Object.entries(models); // Check the models, collect the names of actions in models

    modelEntries.forEach(function (_ref) {
      var _ref2 = _slicedToArray(_ref, 2),
          name = _ref2[0],
          model = _ref2[1];

      checkModel(name, model);
      var actions = model.actions;
      actionNames = [].concat(_toConsumableArray(actionNames), _toConsumableArray(Object.keys(actions)));
    });
    actionNames = _toConsumableArray(new Set(actionNames)); // Check the names of models, create the reducers for models

    modelEntries.forEach(function (_ref3) {
      var _ref4 = _slicedToArray(_ref3, 2),
          name = _ref4[0],
          state = _ref4[1].state;

      if (actionNames.includes(name)) throw new Error(ERR.MODEL_NAME(name));
      rootReducers[name] = createReducer(name, state);
    });
  } else {
    modelEntries = Object.entries(models); // Create the reducers for models

    modelEntries.forEach(function (_ref5) {
      var _ref6 = _slicedToArray(_ref5, 2),
          name = _ref6[0],
          state = _ref6[1].state;

      rootReducers[name] = createReducer(name, state);
    });
  } // Create the Redux store


  var composeEnhancers = useDevTools === true && window.__REDUX_DEVTOOLS_EXTENSION_COMPOSE__ ? window.__REDUX_DEVTOOLS_EXTENSION_COMPOSE__ : redux.compose;
  var store = redux.createStore(redux.combineReducers(rootReducers), composeEnhancers(redux.applyMiddleware.apply(void 0, _toConsumableArray(plugins))));
  var theRealDispatch = store.dispatch;

  store.dispatch = function () {
    throw new Error(ERR.DISPATCH());
  };

  var getState = store.getState,
      dispatch = store.dispatch; // Format the actions of models

  var modelsProxy = {};
  modelEntries.forEach(function (_ref7) {
    var _ref8 = _slicedToArray(_ref7, 1),
        name = _ref8[0];

    modelsProxy[name] = new Proxy({}, createHandler(name, getState, dispatch));
  });
  modelEntries.forEach(function (_ref9) {
    var _ref10 = _slicedToArray(_ref9, 2),
        name = _ref10[0],
        model = _ref10[1];

    var thisProxy = new Proxy({}, createHandler(name, getState, dispatch, modelsProxy));
    createActions(name, model, getState, dispatch, theRealDispatch, thisProxy);
  });

  store.addModel = function (name, model) {
    if (name in rootReducers) return;
    var state = model.state,
        actions = model.actions;

    if (isEnvDevelopment) {
      checkModel(name, model);
      actionNames = _toConsumableArray(new Set([].concat(_toConsumableArray(actionNames), _toConsumableArray(Object.keys(actions)))));
      if (actionNames.includes(name)) throw new Error(ERR.MODEL_NAME(name));
    }

    rootReducers[name] = createReducer(name, state);
    store.replaceReducer(redux.combineReducers(rootReducers));
    modelsProxy[name] = new Proxy({}, createHandler(name, getState, dispatch));
    var thisProxy = new Proxy({}, createHandler(name, getState, dispatch, modelsProxy));
    createActions(name, model, getState, dispatch, theRealDispatch, thisProxy);
  };

  return store;
};
/**
 * Create the mapState and mapActions
 *
 * @param {string[]} names
 * @return {function[]} [mapState, mapActions]
 */


var withStore = function withStore() {
  for (var _len = arguments.length, names = new Array(_len), _key = 0; _key < _len; _key++) {
    names[_key] = arguments[_key];
  }

  var isEnvDevelopment = process.env.NODE_ENV === 'development';
  var allState = {
    loading: {}
  };
  var allActions = {};

  if (isEnvDevelopment) {
    if (names.length === 0) throw new Error(ERR.EMPTY_PARAM());
    var stateChecked = false;
    var actionsChecked = false;
    return [function (state) {
      names.forEach(function (name, index) {
        var partState = state[name];

        if (!stateChecked) {
          if (typeof name !== 'string' || !(name in state)) throw new Error(ERR.PARAM());
          if (index > 0) checkDuplicate(name, 'state', partState, allState);
        }

        allState = _objectSpread$2({}, allState, {}, partState, {}, {
          loading: _objectSpread$2({}, allState.loading, {}, partState.loading)
        });
      });
      if (!stateChecked) stateChecked = true;
      return allState;
    }, function (dispatch) {
      names.forEach(function (name, index) {
        var partActions = dispatch[name];

        if (!actionsChecked) {
          if (index > 0) checkDuplicate(name, 'action', partActions, allActions);
        }

        allActions = _objectSpread$2({}, allActions, {}, partActions);
      });
      if (!actionsChecked) actionsChecked = true;
      return allActions;
    }];
  }

  return [function (state) {
    names.forEach(function (name) {
      var partState = state[name];
      allState = _objectSpread$2({}, allState, {}, partState, {}, {
        loading: _objectSpread$2({}, allState.loading, {}, partState.loading)
      });
    });
    return allState;
  }, function (dispatch) {
    names.forEach(function (name) {
      var partActions = dispatch[name];
      allActions = _objectSpread$2({}, allActions, {}, partActions);
    });
    return allActions;
  }];
};

exports.createStore = createStore;
exports.withStore = withStore;
